/*************************************************************************
*
* VDMIX.C
*
* Virtual Driver Example - Mix
*
* Copyright 1998-2014 Citrix Systems, Inc. All Rights Reserved.
*
* Citrix hereby permits you to use, reproduce, modify, display, compile,
* sublicense, and distribute all or any portion of the sample code contained
* in this file, provided that you remove this Copyright attribution from any
* modified version of this file.
*
* THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
*
*************************************************************************/

#ifdef DOS32
#define CITRIX 1
#include <sys/types.h>
#include "wfc32.h"
#endif

#ifndef unix
#include <windows.h>         /* for far etc. */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "citrix.h"
#include "clib.h"

/* from src/shared/inc/citrix */
#include "ica.h"
#include "ica-c2h.h"
#include "vdmix.h"

/* from src/inc */
#include "icaconst.h"
#include "clterr.h"
#include "wdapi.h"
#include "vdapi.h"
#include "vd.h"
#include "ctxdebug.h"
#include "logapi.h"
#include "miapi.h"


#ifdef unix
// make sure that we pull in Load from vdapi.a
extern int Load();

struct _wrapload {
        int ((*fn)());
} WRAP_LOAD[] = {
        {Load}
};
#endif	// unix

/*=============================================================================
==   Functions Defined
=============================================================================*/
/*
 *   These are the required driver functions (exported)
 */

int DriverOpen( PVD, PVDOPEN, PUINT16 );
int DriverClose( PVD, PDLLCLOSE, PUINT16 );
int DriverInfo( PVD, PDLLINFO, PUINT16 );
int DriverPoll( PVD, PDLLPOLL, PUINT16 );
int DriverQueryInformation( PVD, PVDQUERYINFORMATION, PUINT16 );
int DriverSetInformation( PVD, PVDSETINFORMATION, PUINT16 );
int DriverGetLastError( PVD, PVDLASTERROR );

static int WFCAPI ICADataArrival( PVOID, USHORT, LPBYTE, USHORT );


/*=============================================================================
==   Data
=============================================================================*/


/*
 *  These are returned when we register our hook
 */

PVOID pWd = NULL;
POUTBUFRESERVEPROC  pOutBufReserve  = NULL;
POUTBUFAPPENDPROC   pOutBufAppend   = NULL;
POUTBUFWRITEPROC    pOutBufWrite    = NULL;
PAPPENDVDHEADERPROC pAppendVdHeader = NULL;

USHORT      fBufferEmpty = TRUE;          /* True if the data buffer is empty */
PMIXHEAD    pMixHd = NULL ;               /* pointer to mix data buffer */
LPBYTE      pTemp = NULL;                 /* Pointer for temporary buffer */
static      USHORT gusMaxDataSize;        /* Maximum Data Write Size */
static      USHORT gusVirtualChannelNum;  /* Channel number assigned by WD */


/*******************************************************************************
 *
 *  DriverOpen
 *
 *    Called once to set up things.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdOpen (input/output)
 *       pointer to the structure VDOPEN
 *    puiSize (Output)
 *       size of VDOPEN structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *    CLIENT_ERROR_NO_MEMORY - could not allocate data buffer
 *    On other errors, an error code is returned from lower level functions.
 *
 ******************************************************************************/

int
DriverOpen( PVD pVd, PVDOPEN pVdOpen, PUINT16 puiSize )
{
    WDSETINFORMATION   wdsi;
    VDWRITEHOOK        vdwh;
    WDQUERYINFORMATION wdqi;
    OPENVIRTUALCHANNEL OpenVirtualChannel;
    int                rc;
    UINT16             uiSize;

    TRACE(( TC_VD, TT_API1, "VDMIX: DriverOpen entered" ));

    fBufferEmpty = TRUE;

    *puiSize = sizeof(VDOPEN);

    /*
     * Get a virtual channel
     */
    wdqi.WdInformationClass = WdOpenVirtualChannel;
    wdqi.pWdInformation = &OpenVirtualChannel;
    wdqi.WdInformationLength = sizeof(OPENVIRTUALCHANNEL);
    OpenVirtualChannel.pVCName = CTXMIX_VIRTUAL_CHANNEL_NAME;

    /* uiSize will be set  when we return back from VdCallWd. */
    uiSize = sizeof( WDQUERYINFORMATION);

    rc = VdCallWd( pVd, WDxQUERYINFORMATION, &wdqi, &uiSize );
    TRACE(( TC_VD, TT_API1, "VDMIX: opening channel %s",
            OpenVirtualChannel.pVCName ));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE(( TC_VD, TT_ERROR, "VDMIX: Could not open %s. rc=%d.",
                OpenVirtualChannel.pVCName, rc));
        return( rc );
    }

    gusVirtualChannelNum = OpenVirtualChannel.Channel;

    /*
     *  Used the by engine
     */
    pVdOpen->ChannelMask = (1L << gusVirtualChannelNum);
    pVd->pPrivate   = NULL; /* pointer to private data, if needed */

    /*
     *  Register write hooks for our virtual channel
     */
    vdwh.Type  = gusVirtualChannelNum;
    vdwh.pVdData = pVd;
    vdwh.pProc = ICADataArrival;
    wdsi.WdInformationClass  = WdVirtualWriteHook;
    wdsi.pWdInformation      = &vdwh;
    wdsi.WdInformationLength = sizeof(VDWRITEHOOK);
    uiSize                   = sizeof( WDSETINFORMATION);

    rc = VdCallWd( pVd, WDxSETINFORMATION, &wdsi, &uiSize );
    TRACE(( TC_VD, TT_API2, "VDMIX: writehook channel number=%u"
            " vdwh.pVdData=%x rc=%d", vdwh.Type, vdwh.pVdData, rc ));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE(( TC_VD, TT_ERROR, "VDMIX: Could not register write hook. rc %d",
                rc));
        return( rc );
    }

    /*
     *  Grab pointers to functions to use to send data to the host
     */
    pWd            = vdwh.pWdData;
    pOutBufReserve = vdwh.pOutBufReserveProc;
    pOutBufAppend  = vdwh.pOutBufAppendProc;
    pOutBufWrite   = vdwh.pOutBufWriteProc;
    pAppendVdHeader= vdwh.pAppendVdHeaderProc;

    TRACE(( TC_VD, TT_API1, "VDMIX: Registered"));

    /*
     *  ALL MEMORY MUST BE ALLOCATED DURING INITIALIZATION.
     *  Allocate a single buffer to respond to the mix request.
     *  This example shows use of the MaximumWriteSize returned via
     *  the previous call.
     *  Subtract one because the first byte is used internally by the
     *  WD for the channel number.  We are given a pointer to the
     *  second byte.
     */
    gusMaxDataSize = vdwh.MaximumWriteSize - 1;
    /*
     *  This is the main buffer which is going to be exchanged
     *  between the client and the server. We allocate memory
     *  of size gusMaxDataSize rather than sizeof(MIXHEAD)
     *  as the same buffer is used to send back additional data
     *  to the server if required
     */
    if ( (pMixHd = (PMIXHEAD)malloc(gusMaxDataSize)) == NULL ) {
        return( CLIENT_ERROR_NO_MEMORY );
    }

    /*
     *  This is a buffer to write the temporary display string
     * in the DispStr function to the Log file
     */
    if ( (pTemp = (LPBYTE)malloc(TEMP_SIZE)) == NULL ) {
        return( CLIENT_ERROR_NO_MEMORY );
    }

    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  ICADataArrival
 *
 *   A data PDU arrived over our channel.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *
 *    uChan (input)
 *       ICA channel the data is for.
 *
 *    pBuf (input)
 *       Buffer with arriving data packet
 *
 *    Length (input)
 *       Length of the data packet
 *
 * EXIT:
 *        CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

static int WFCAPI ICADataArrival( PVOID pVd, USHORT uChan, LPBYTE pBuf, USHORT Length)
{
    TRACE(( TC_VD, TT_API3,
       "VDMIX: ICADataArrival - Entered "));

    TRACE(( TC_VD, TT_API3,
           "VDMIX: IcaDataArrival, Len=%d, uLen=%d, pBuf=[%x][%x][%x][%x]",
           Length, ((PMIXHEAD)pBuf)->uLen, pBuf[0], pBuf[1], pBuf[2], pBuf[3]));
    TRACEBUF(( TC_VD, TT_API3, pBuf, Length));

    /*
     *  This protocol is completely synchronous - host should not send
     *  another message with a pending response.
     */
    if ( !fBufferEmpty ) {
        TRACE(( TC_VD, TT_ERROR,
            "VDMIX: ICADataArrival - Error: not all data was sent" ));
        return CLIENT_STATUS_SUCCESS;
    }

    /*
     *  Fill in the header first
     */
    memcpy( pMixHd, pBuf, sizeof(MIXHEAD));
    pBuf +=  sizeof(MIXHEAD);

    /*
     *  Call the appropriate function as per the uFunc field in the Header
     */
    TRACE(( TC_VD, TT_API3, "Function received is %d", pMixHd->uFunc));

    if (pMixHd->uFunc==ADD)
        AddNo(pBuf);
    if (pMixHd->uFunc==DISP)
        DispStr(pBuf);
    if (pMixHd->uFunc==TIME)
        LocTime(pBuf);

    /*
     *   If a return value is requested, fBufferEmpty Flag is set to False
     *   This will indicate the DriverPoll function that it needs to send
     *   a response back to the server.
     */

    if (pMixHd->fRetReq) {
        fBufferEmpty = FALSE;

    TRACE(( TC_VD, TT_ERROR,
           "VDMIX: fBufferEmpty flag set to False"));
    }

    TRACEBUF(( TC_VD, TT_API3, pMixHd, 12 ));
    return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverPoll
 *
 *  The Winstation driver calls DriverPoll
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdPoll (input)
 *       pointer to the structure DLLPOLL  (not used)
 *    puiSize (input)
 *       size of DLLPOOL structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - OK so far.  We will be polled again.
 *    CLIENT_STATUS_NO_DATA - No more data to send.
 *    CLIENT_STATUS_ERROR_RETRY - Could not get an output buffer.
 *                            Hopefully, one will be available next time.
 *    Otherwise, a fatal error code is returned from lower level functions.
 *
 *    NOTE:  CLIENT_STATUS_NO_DATA signals the WD that it is OK to stop
 *           polling until another host to client packet comes through.
 ******************************************************************************/

int
DriverPoll( PVD pVd, PDLLPOLL pVdPoll, PUINT16 puiSize )
{
    int rc = CLIENT_STATUS_NO_DATA;

    static USHORT fFirstTimeDebug = TRUE;  /* Only print on first invocation */

    if ( fFirstTimeDebug == TRUE ) {
        TRACE(( TC_VD, TT_API2, "VDMIX: DriverPoll entered" ));
        fFirstTimeDebug = FALSE;
    }

    /*
     *  Check for something to write
     */
    if ( fBufferEmpty ) {
        rc = CLIENT_STATUS_NO_DATA;
        goto Exit;
    }

    /*
     *  Write is postponed until this poll loop because the outbufreserve can
     *  fail, forcing us to wait until a buffer is freed - the only way to wait
     *  is to pick things up during the poll loop.
     *
     *  Reserve output buffer space; fudge the length for the ICA virtual
     *  header overhead.
     */
    if ( pOutBufReserve( pWd, (USHORT)(pMixHd->uLen+4) ) ) {
        rc =  CLIENT_STATUS_ERROR_RETRY;
        goto Exit;
    }

    /*
     *  Append ICA Virtual write header
     */
    if (( rc = pAppendVdHeader( pWd, gusVirtualChannelNum, (USHORT)pMixHd->uLen ))
               != CLIENT_STATUS_SUCCESS ) {
        goto Exit;
    }

    /*
     *  Append the mix packet
     */
    if (( rc = pOutBufAppend( pWd, (LPBYTE)pMixHd, (USHORT)pMixHd->uLen ))
               != CLIENT_STATUS_SUCCESS ) {
        goto Exit;
    }

    /*
     *  Write the buffer
     */
    if (( rc = pOutBufWrite( pWd ))
               != CLIENT_STATUS_SUCCESS ) {
        goto Exit;
    }

    fBufferEmpty = TRUE;

    TRACE(( TC_VD, TT_ERROR,
        "VDMIX: fBufferEmpty flag set to TRUE"));
    rc = CLIENT_STATUS_SUCCESS;

Exit:
    return( rc );
}


/*******************************************************************************
 *
 *  DriverClose
 *
 *  The user interface calls VdClose to close a Vd before unloading it.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to procotol driver data structure
 *    pVdClose (input/output)
 *       pointer to the structure DLLCLOSE
 *    puiSize (input)
 *       size of DLLCLOSE structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int
DriverClose( PVD pVd, PDLLCLOSE pVdClose, PUINT16 puiSize )
{

    TRACE(( TC_VD, TT_API1, "VDMIX: DriverClose entered" ));

    free( pMixHd );
    free(pTemp);
    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverInfo
 *
 *    This routine is called to get module information
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdInfo (output)
 *       pointer to the structure DLLINFO
 *    puiSize (output)
 *       size of DLLINFO structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverInfo( PVD pVd, PDLLINFO pVdInfo, PUINT16 puiSize )
{
    USHORT ByteCount;
    PVDMIX_C2H pVdData;
    PMODULE_C2H pHeader;
    PVDFLOW pFlow;

    ByteCount = sizeof(VDMIX_C2H);

    *puiSize = sizeof(DLLINFO);

    TRACE(( TC_VD, TT_API1, "VDMIX: DriverInfo entered" ));

    /*
     *  Check if buffer is big enough
     *  If not, the caller is probably trying to determine the required
     *  buffer size, so return it in ByteCount.
     */
    if ( pVdInfo->ByteCount < ByteCount ) {
        pVdInfo->ByteCount = ByteCount;
        return( CLIENT_ERROR_BUFFER_TOO_SMALL );
    }

    /*
     *  Initialize default data
     */
    pVdInfo->ByteCount = ByteCount;
    pVdData = (PVDMIX_C2H) pVdInfo->pBuffer;

    /*
     *  Initialize module header
     */
    pHeader = &pVdData->Header.Header;
    pHeader->ByteCount = ByteCount;
    pHeader->ModuleClass = Module_VirtualDriver;

    pHeader->VersionL = CTXMIX_VER_LO;
    pHeader->VersionH = CTXMIX_VER_HI;
    strcpy((char*)(pHeader->HostModuleName), "ICA"); /* max 8 characters */

    /*
     *  Initialize virtual driver header
     */
    pVdData->Header.ChannelMask = pVd->ChannelMask;
    pFlow = &pVdData->Header.Flow;
    pFlow->BandwidthQuota = 0;
    pFlow->Flow = VirtualFlow_None;

    /* add our own data */
    pVdData->usMaxDataSize = gusMaxDataSize;
    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverQueryInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdQueryInformation (input/output)
 *       pointer to the structure VDQUERYINFORMATION
 *    puiSize (output)
 *       size of VDQUERYINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverQueryInformation( PVD pVd, PVDQUERYINFORMATION pVdQueryInformation, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDMIX: DriverQueryInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdQueryInformation->VdInformationClass = %d",
         pVdQueryInformation->VdInformationClass));

    *puiSize = sizeof( VDQUERYINFORMATION );

    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverSetInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdSetInformation (input/output)
 *       pointer to the structure VDSETINFORMATION
 *    puiSize (input)
 *       size of VDSETINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverSetInformation( PVD pVd, PVDSETINFORMATION pVdSetInformation, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDMIX: DriverSetInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdSetInformation->VdInformationClass = %d",
         pVdSetInformation->VdInformationClass));
    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverGetLastError
 *
 *   Queries error data.
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pLastError (output)
 *       pointer to the error structure to return (message is currently always
 *       NULL)
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverGetLastError( PVD pVd, PVDLASTERROR pLastError )
{
    TRACE(( TC_VD, TT_API1, "VDMIX: DriverGetLastError entered" ));
    /*
     *  Interpret last error and pass back code and string data
     */

    pLastError->Error = pVd->LastError;
    pLastError->Message[0] = '\0';
    return(CLIENT_STATUS_SUCCESS);
}


/******************************************************************************
 *
 *  ADDNO:
 *         Function executed on client side to Add two Numbers. Example
 *         of a function that requires a return Value.
 *
 *  ENTRY: pBufIp
 *           Pointer to the Arguments to the function
 *
 *  EXIT:  -
 *
 ******************************************************************************/

VOID
AddNo(LPBYTE pBufIp)
{
    USHORT num1;
    USHORT num2;

    TRACE(( TC_VD, TT_API3, "VDMIX: ADDNO"));
    memcpy(&num1, pBufIp, sizeof(USHORT));
    pBufIp += sizeof(USHORT);
    memcpy(&num2, pBufIp, sizeof(USHORT));
    pMixHd->uLen = sizeof(MIXHEAD);
    TRACE(( TC_VD, TT_API3, "VDMIX:Addno %d  %d  ", num1, num2));
    pMixHd->dwRetVal =  num1+ num2;
    TRACE(( TC_VD, TT_API2, "Result of addition is %d", pMixHd->dwRetVal));
    return;
}

/******************************************************************************
 *
 *  DispStr:
 *         Function executed on client side to Display a String
 *         (It is written to a log file) Example of a function that
 *         does not require a return value
 *
 *  ENTRY: pBufIp
 *           Pointer to the Arguments to the function
 *
 *  EXIT:  -
 *
 ******************************************************************************/

VOID
DispStr(LPBYTE pBufIp)
{
    memset(pTemp, '\0', TEMP_SIZE);
    memcpy(pTemp, pBufIp, pMixHd->dwLen1);
    TRACE(( TC_VD, TT_API1, "The String Received is '%s'", pTemp));
    printf("DispStr:%s\n", pTemp);
}


/******************************************************************************
 *
 *  LocTime:
 *         Function returns the local time on the executed on client side
 *
 *  ENTRY: pBufIp
 *           Pointer to the Arguments to the function
 *
 *  EXIT:  -
 *
 ******************************************************************************/

VOID
LocTime(LPBYTE pBufIp)
{
    LPBYTE pData;
    pData = (LPBYTE)pMixHd;
    pData += sizeof(MIXHEAD);
    TRACE(( TC_VD, TT_API2, "VDMIX: Get Local Time"));

#if !defined(DOS32) && !defined(WINCE)
    {
    struct tm *newtime;
    VDTM loctm;
    time_t long_time;

    time( &long_time );                /* Get time as long integer. */
    newtime = localtime( &long_time ); /* Convert to local time. */

    loctm.tm_sec = newtime->tm_sec;     /* seconds after the minute - [0,59] */
    loctm.tm_min = newtime->tm_min;     /* minutes after the hour - [0,59] */
    loctm.tm_hour = newtime->tm_hour;   /* hours since midnight - [0,23] */
    loctm.tm_mday = newtime->tm_mday;   /* day of the month - [1,31] */
    loctm.tm_mon = newtime->tm_mon;     /* months since January - [0,11] */
    loctm.tm_year = newtime->tm_year;   /* years since 1900 */
    loctm.tm_wday = newtime->tm_wday;   /* days since Sunday - [0,6] */
    loctm.tm_yday = newtime->tm_yday;   /* days since January 1 - [0,365] */
    loctm.tm_isdst = newtime->tm_isdst; /* daylight savings time flag */

    memcpy(pData, &loctm, sizeof(VDTM));
    pMixHd->uLen = sizeof(MIXHEAD) + sizeof(VDTM);
    pMixHd->dwRetVal =  sizeof(MIXHEAD);
    TRACE(( TC_VD, TT_API2,
        "The two lengths are (total)%d (mixhead)%d",
        pMixHd->uLen, pMixHd->dwRetVal));
    TRACE(( TC_VD, TT_API2, "Local Time is %s", asctime( (struct tm*)pData )));
    }
#else
    /*
     *  This example does not return time for the DOS32 client. Value 1001
     * (arbitarily chosen) is used to convey this
     */
    pMixHd->dwRetVal =  1001;
    pMixHd->uLen = sizeof(MIXHEAD);
#endif

    return;
}

