/*************************************************************************
*
* VDOVER.C
*
* Virtual Driver Example - Over
*
*  Copyright 1998-2014 Citrix Systems, Inc. All Rights Reserved.
*
* Citrix hereby permits you to use, reproduce, modify, display, compile,
* sublicense, and distribute all or any portion of the sample code contained
* in this file, provided that you remove this Copyright attribution from any
* modified version of this file.
*
* THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
*
*************************************************************************/

#ifdef DOS32
#define CITRIX 1
#include <sys/types.h>
#include "wfc32.h"
#endif

#ifndef unix
#include <windows.h>         /* for far etc. */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "citrix.h"
#include "clib.h"

/* from src/shared/inc/citrix */
#include "ica.h"
#include "ica-c2h.h"
#include "vdover.h"

/* from src/inc */
#include "icaconst.h"
#include "clterr.h"
#include "wdapi.h"
#include "vdapi.h"
#include "vd.h"
#include "ctxdebug.h"
#include "logapi.h"
#include "miapi.h"

#ifdef unix
// make sure that we pull in Load from vdapi.a
extern int Load();

struct _wrapload {
        int ((*fn)());
} WRAP_LOAD[] = {
        {Load}
};
#endif	// unix

/*=============================================================================
==   Functions Defined
=============================================================================*/
/*
 *   These are the required driver functions (exported)
 */

int DriverOpen( PVD, PVDOPEN, PUINT16 );
int DriverClose( PVD, PDLLCLOSE, PUINT16 );
int DriverInfo( PVD, PDLLINFO, PUINT16 );
int DriverPoll( PVD, PDLLPOLL, PUINT16 );
int DriverQueryInformation( PVD, PVDQUERYINFORMATION, PUINT16 );
int DriverSetInformation( PVD, PVDSETINFORMATION, PUINT16 );
int DriverGetLastError( PVD, PVDLASTERROR );

static int WFCAPI ICADataArrival( PVOID, USHORT, LPBYTE, USHORT );

/*=============================================================================
==   Data
=============================================================================*/


/*
 *  These are returned when we register our hook
 */

PVOID pWd = NULL;
POUTBUFRESERVEPROC  pOutBufReserve  = NULL;
POUTBUFAPPENDPROC   pOutBufAppend   = NULL;
POUTBUFWRITEPROC    pOutBufWrite    = NULL;
PAPPENDVDHEADERPROC pAppendVdHeader = NULL;


BOOL fBufferEmpty = TRUE;        /* True if the data buffer is empty */
PDRVRESP pDrvToSrv;              /* pointer to reply buffer from drv to Srv */
POVER    pOver;                  /* Pointer to receive data from Srv */

static USHORT gusMaxDataSize;             /* Maximum Data Write Size */
static USHORT gusVirtualChannelNum;       /* Channel number assigned by WD */


/*******************************************************************************
 *
 *  DriverOpen
 *
 *    Called once to set up things.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdOpen (input/output)
 *       pointer to the structure VDOPEN
 *    puiSize (Output)
 *       size of VDOPEN structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *    CLIENT_ERROR_NO_MEMORY - could not allocate data buffer
 *    On other errors, an error code is returned from lower level functions.
 *
 ******************************************************************************/

int
DriverOpen( PVD pVd, PVDOPEN pVdOpen, PUINT16 puiSize )
{
    WDSETINFORMATION   wdsi;
    VDWRITEHOOK        vdwh;
    WDQUERYINFORMATION wdqi;
    OPENVIRTUALCHANNEL OpenVirtualChannel;
    int                rc;
    UINT16             uiSize;

    TRACE(( TC_VD, TT_API1, "VDOVER: DriverOpen entered" ));

    fBufferEmpty = TRUE;

    *puiSize = sizeof(VDOPEN);

    /*
     * Get a virtual channel
     */
    wdqi.WdInformationClass = WdOpenVirtualChannel;
    wdqi.pWdInformation = &OpenVirtualChannel;
    wdqi.WdInformationLength = sizeof(OPENVIRTUALCHANNEL);
    OpenVirtualChannel.pVCName = CTXOVER_VIRTUAL_CHANNEL_NAME;

    /* uiSize will be set  when we return back from VdCallWd. */
    uiSize = sizeof( WDQUERYINFORMATION);

    rc = VdCallWd( pVd, WDxQUERYINFORMATION, &wdqi, &uiSize  );
    TRACE(( TC_VD, TT_API1, "VDOVER: opening channel %s",
            OpenVirtualChannel.pVCName ));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE(( TC_VD, TT_ERROR, "VDOVER: Could not open %s. rc=%d.",
                OpenVirtualChannel.pVCName, rc));
        return( rc );
    }

    gusVirtualChannelNum = OpenVirtualChannel.Channel;

    /*
     *  Used by the engine
     */
    pVdOpen->ChannelMask = (1L << gusVirtualChannelNum);
    pVd->pPrivate   = NULL; /* pointer to private data, if needed */

    /*
     *  Register write hooks for our virtual channel
     */
    vdwh.Type  = gusVirtualChannelNum;
    vdwh.pVdData = pVd;
    vdwh.pProc = ICADataArrival;
    wdsi.WdInformationClass  = WdVirtualWriteHook;
    wdsi.pWdInformation      = &vdwh;
    wdsi.WdInformationLength = sizeof(VDWRITEHOOK);
    uiSize                   = sizeof( WDSETINFORMATION);

    rc = VdCallWd( pVd, WDxSETINFORMATION, &wdsi, &uiSize  );
    TRACE(( TC_VD, TT_API2, "VDOVER: writehook channel number=%u"
            " vdwh.pVdData=%x rc=%d",
            vdwh.Type, vdwh.pVdData, rc ));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE(( TC_VD, TT_ERROR, "VDOVER: Could not register write hook. rc %d",
                rc));
        return( rc );
    }

    /*
     *  Grab pointers to functions to use to send data to the host
     */
    pWd            = vdwh.pWdData;
    pOutBufReserve = vdwh.pOutBufReserveProc;
    pOutBufAppend  = vdwh.pOutBufAppendProc;
    pOutBufWrite   = vdwh.pOutBufWriteProc;
    pAppendVdHeader= vdwh.pAppendVdHeaderProc;

    TRACE(( TC_VD, TT_API1, "VDOVER: Registered"));

    /*
     *  ALL MEMORY MUST BE ALLOCATED DURING INITIALIZATION.
     *  Allocate a single buffer to send commands to the server
     *  This example shows use of the MaximumWriteSize returned via
     *  the previous call.
     *  Subtract one because the first byte is used internally by the
     *  WD for the channel number.  We are given a pointer to the
     *  second byte.
     */
    if ( (pDrvToSrv = (PDRVRESP)malloc(sizeof(DRVRESP) )) == NULL ) {
        return( CLIENT_ERROR_NO_MEMORY );
    }

    memset(pDrvToSrv, 0, sizeof(DRVRESP));

    /*
     * pOver is used as the buffer to receive the data from the server
     */

    if ( (pOver = (POVER)malloc(sizeof(OVER))) == NULL ) {
        return( CLIENT_ERROR_NO_MEMORY );
    }
    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  ICADataArrival
 *
 *   A data PDU arrived over our channel.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *
 *    uChan (input)
 *       ICA channel the data is for.
 *
 *    pBuf (input)
 *       Buffer with arriving data packet
 *
 *    Length (input)
 *       Length of the data packet
 *
 * EXIT:
 *        CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

static int WFCAPI ICADataArrival( PVOID pVd, USHORT uChan, LPBYTE pBuf,
                                  USHORT Length )
{
    /*
     * ICADataArrival is called (polled) when the underlying WinStation 
     * Driver(WD) has data to pass to the Virtual Channel.This fBufferEmpty 
     * condition is added so that the client does not accept any more data till
     * it has send back a message to the server.The data send by the server 
     * during this time will be lost. If the following condition is not checked 
     * then the client can continue to receive data from the server.(The data 
     * being stored by the WD.)  
     */
     
    if ( !fBufferEmpty ) {
        TRACE(( TC_VD, TT_ERROR,
            "VDOVER: ICADataArrival - Error: not all data was sent" ));
        return CLIENT_STATUS_SUCCESS;
    }

    TRACE(( TC_VD, TT_API3,
        "VDOVER: IcaDataArrival, Len=%d, uLen=%d, ulServerMS=%d",
        Length, ((POVER)pBuf)->uLen, ((POVER)pBuf)->ulServerMS ));
    TRACEBUF(( TC_VD, TT_API3, pBuf, Length));


    if (((POVER)pBuf)->uLen > gusMaxDataSize) {
        pOver->uLen = gusMaxDataSize;
    }
    else {
        pOver->uLen = ((POVER)pBuf)->uLen;
    }

    /*
     * Copy data from pBuf - buffer with arriving data packet to
     * pOver - buffer to hold the packet type being sent by the
     * server.
     */
    memcpy( pOver, pBuf, sizeof(OVER) );
    /*
     *  When the counter is divisible by 10, we increment it by 7
     *  Numbers 10 and 7 are chosen arbitarily
     */
    if (!((pOver->uCounter) % 10)) {
        fBufferEmpty=FALSE;
        TRACE(( TC_VD, TT_ERROR,
          "VDOVER: ICADataArrival fBufferEmpty made false" ));
        pDrvToSrv->uCounter = pOver->uCounter + 7;
    }

    TRACE(( TC_VD, TT_API3,
        "VDOVER: ICADataArrival uCounter %d ", pOver->uCounter ));

    return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverPoll
 *
 *  The Winstation driver calls DriverPoll
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdPoll (input)
 *       pointer to the structure DLLPOLL  (not used)
 *    puiSize (input)
 *       size of DLLPOOL structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - OK so far.  We will be polled again.
 *    CLIENT_STATUS_NO_DATA - No more data to send.
 *    CLIENT_STATUS_ERROR_RETRY - Could not get an output buffer.
 *                            Hopefully, one will be available next time.
 *    Otherwise, a fatal error code is returned from lower level functions.
 *
 *    NOTE:  CLIENT_STATUS_NO_DATA signals the WD that it is OK to stop
 *           polling until another host to client packet comes through.
 ******************************************************************************/

int
DriverPoll( PVD pVd, PDLLPOLL pVdPoll, PUINT16 puiSize )
{
    int rc = CLIENT_STATUS_NO_DATA;

    static BOOL fFirstTimeDebug = TRUE;  /* Only print on first invocation */

    if ( fFirstTimeDebug == TRUE ) {
        TRACE(( TC_VD, TT_API2, "VDOVER: DriverPoll entered" ));
        fFirstTimeDebug = FALSE;
    }

    /*
     *  Check for something to write
     */
    if ( fBufferEmpty ) {
        rc = CLIENT_STATUS_NO_DATA;
        goto Exit;
    }

     pDrvToSrv->uType = OVERFLOW_JUMP;
     pDrvToSrv->uLen = sizeof(DRVRESP);

    /*
     *  Write is postponed until this poll loop because the outbufreserve can
     *  fail, forcing us to wait until a buffer is freed - the only way to wait
     *  is to pick things up during the poll loop.
     *
     *  Reserve output buffer space; fudge the length for the ICA virtual
     *  header overhead.
     */
    if ( pOutBufReserve( pWd, (USHORT)(pDrvToSrv->uLen+4) ) ) {
        rc =  CLIENT_STATUS_ERROR_RETRY;
        goto Exit;
    }



    /*
     *  Fill in Client to Server packet
     */
    TRACE(( TC_VD, TT_API2, "VDOVER: Buffer to Write, type=%u",
        pDrvToSrv->uType ));

    /*
     *  Append ICA Virtual write header
     */
    if (( rc = pAppendVdHeader( pWd, gusVirtualChannelNum, pDrvToSrv->uLen ))
               != CLIENT_STATUS_SUCCESS ) {
        goto Exit;
    }

    /*
     *  Append the reply packet
     */
    if (( rc = pOutBufAppend( pWd, (LPBYTE)pDrvToSrv, pDrvToSrv->uLen ))
               != CLIENT_STATUS_SUCCESS ) {
        goto Exit;
    }

    /*
     *  Write the buffer
     */
    if (( rc = pOutBufWrite( pWd ))
               != CLIENT_STATUS_SUCCESS ) {
        goto Exit;
    }

    fBufferEmpty = TRUE;
    TRACE(( TC_VD, TT_API2, "VDOVER: fBufferEmpty made TRUE"));

    rc = CLIENT_STATUS_SUCCESS;

Exit:
    return( rc );
}


/*******************************************************************************
 *
 *  DriverClose
 *
 *  The user interface calls VdClose to close a Vd before unloading it.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to procotol driver data structure
 *    pVdClose (input/output)
 *       pointer to the structure DLLCLOSE
 *    puiSize (input)
 *       size of DLLCLOSE structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int
DriverClose( PVD pVd, PDLLCLOSE pVdClose, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDOVER: DriverClose entered" ));
    if (pDrvToSrv)
        free( pDrvToSrv );
    if (pOver)
        free(pOver);
    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverInfo
 *
 *    This routine is called to get module information
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdInfo (output)
 *       pointer to the structure DLLINFO
 *    puiSize (output)
 *       size of DLLINFO structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverInfo( PVD pVd, PDLLINFO pVdInfo, PUINT16 puiSize )
{
    USHORT ByteCount;
    PVDOVER_C2H pVdData;
    PMODULE_C2H pHeader;
    PVDFLOW pFlow;

    ByteCount = sizeof(VDOVER_C2H);

    *puiSize = sizeof(DLLINFO);

    TRACE(( TC_VD, TT_API1, "VDOVER: DriverInfo entered" ));

    /*
     *  Check if buffer is big enough
     *  If not, the caller is probably trying to determine the required
     *  buffer size, so return it in ByteCount.
     */
    if ( pVdInfo->ByteCount < ByteCount ) {
        pVdInfo->ByteCount = ByteCount;
        return( CLIENT_ERROR_BUFFER_TOO_SMALL );
    }

    /*
     *  Initialize default data
     */
    pVdInfo->ByteCount = ByteCount;
    pVdData = (PVDOVER_C2H) pVdInfo->pBuffer;

    /*
     *  Initialize module header
     */
    pHeader = &pVdData->Header.Header;
    pHeader->ByteCount = ByteCount;
    pHeader->ModuleClass = Module_VirtualDriver;

    pHeader->VersionL = CTXOVER_VER_LO;
    pHeader->VersionH = CTXOVER_VER_HI;
    strcpy((char*)(pHeader->HostModuleName), "ICA"); /* max 8 characters */

    /*
     *  Initialize virtual driver header
     */
    pVdData->Header.ChannelMask = pVd->ChannelMask;
    pFlow = &pVdData->Header.Flow;
    pFlow->BandwidthQuota = 0;
    pFlow->Flow = VirtualFlow_None;

    /* add our own data */
    pVdData->usMaxDataSize = gusMaxDataSize;

    return( CLIENT_STATUS_SUCCESS );
}



/*******************************************************************************
 *
 *  DriverQueryInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdQueryInformation (input/output)
 *       pointer to the structure VDQUERYINFORMATION
 *    puiSize (output)
 *       size of VDQUERYINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverQueryInformation( PVD pVd, PVDQUERYINFORMATION pVdQueryInformation, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDOVER: DriverQueryInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdQueryInformation->VdInformationClass = %d",
         pVdQueryInformation->VdInformationClass));

    *puiSize = sizeof( VDQUERYINFORMATION );

    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverSetInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdSetInformation (input/output)
 *       pointer to the structure VDSETINFORMATION
 *    puiSize (input)
 *       size of VDSETINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverSetInformation( PVD pVd, PVDSETINFORMATION pVdSetInformation, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDOVER: DriverSetInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdSetInformation->VdInformationClass = %d",
         pVdSetInformation->VdInformationClass));

   return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverGetLastError
 *
 *   Queries error data.
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pLastError (output)
 *       pointer to the error structure to return (message is currently always
 *       NULL)
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverGetLastError( PVD pVd, PVDLASTERROR pLastError )
{
    TRACE(( TC_VD, TT_API1, "VDOVER: DriverGetLastError entered" ));
    /*
     *  Interpret last error and pass back code and string data
     */

    pLastError->Error = pVd->LastError;
    pLastError->Message[0] = '\0';
    return(CLIENT_STATUS_SUCCESS);
}

